local epsCastbars = CreateFrame("Frame", "epsCastbars", UIParent)
_G["epsCastbars"] = epsCastbars

function epsCastbars:round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function epsCastbars:HideBlizzard(unit)
    if(unit == "player") then
        CastingBarFrame:UnregisterAllEvents()
        CastingBarFrame.Show = function() end
        CastingBarFrame:Hide()
    end
end

function epsCastbars:CastInfo(unit)
    local spell, rank, name, icon, startTime, endTime, isTradeSkill = UnitCastingInfo(unit)
    local channeled = false
    
    if(not spell) then
        channeled = true
        spell, rank, name, icon, startTime, endTime, isTradeSkill = UnitChannelInfo(unit)
    end

    return spell, rank, name, icon, startTime, endTime, isTradeSkill, channeled
end

function epsCastbars:IsCasting(unit)
    local spell = select(1, self:CastInfo(unit))
    
    return spell ~= nil and true or false
end

function epsCastbars:Start(unit)
    local spell, rank, name, icon, startTime, endTime, isTradeSkill, channeled = self:CastInfo(unit)
    
    local f = self[unit]
    
    f.startTime = startTime
    f.endTime = endTime
    f.duration = endTime-startTime
    f.initialDuration = f.duration
    f.channeled = channeled
    
    f.name:SetText(spell)
    f.icon:SetTexture(icon)
    
    if(channeled) then
        f.progress:SetStatusBarColor(1, 122/255, 0, 1)
    else
        f.progress:SetStatusBarColor(0, 122/255, 1, 1)
    end
    
    f.fade = 0
    f.fadeAlpha = 150
    f:SetAlpha(1)
    f:Show()
end

function epsCastbars:Stop(unit)
    local f = self[unit]
    local success = false
    
    if(f.endTime) then
        local lag = select(3, GetNetStats())
        local eta = (GetTime()*1000)-f.endTime
        
        if(eta+lag >= 0) then
            success = true
        end
    end
    
    f.startTime = nil
    f.endTime = nil
    f.duration = nil
    f.initialDuration = nil
    
    if(not success) then
        f.progress:SetStatusBarColor(122/255, 0, 0, 1)
    end
    
    self:FadeOut(unit)
end

function epsCastbars:FadeOut(unit)
    local f = self[unit]
    
    f.fade = 1
end

function epsCastbars:Spawn(unit)
    self:HideBlizzard(unit)
    
    local f = CreateFrame("Frame", nil, epsCastbars)
    self[unit] = f
    f.unit = unit
    f.lingerDuration = .5
    
    f:SetWidth(200)
    f:SetHeight(32)
    f:SetBackdrop({
        bgFile = "Interface\\AddOns\\epsCastbars\\background", tile = true, tileSize = 16,
        edgeFile = "Interface\\AddOns\\epsCastbars\\border", edgeSize = 16,
    })
    f:SetBackdropBorderColor(10/255, 10/255, 10/255, 1)
    f:SetBackdropColor(1, 1, 1, 0)
    f:Hide()
    
    f.icon = f:CreateTexture()
    f.icon:SetPoint("TOPLEFT", f, "TOPLEFT", 5, -5)
    f.icon:SetWidth(f:GetHeight()-10)
    f.icon:SetHeight(f:GetHeight()-10)
    f.icon:SetTexCoord(0.1, .9, 0.1, .9)
    
    f.progress = CreateFrame("StatusBar", nil, f)
    f.progress:SetStatusBarTexture("Interface\\AddOns\\epsCastbars\\foreground_healthbar")
    f.progress:SetBackdrop({
        bgFile = "Interface\\AddOns\\epsCastbars\\background",
        tile = true
    })
    f.progress:SetBackdropColor(.5, .5, .5, .5)
    f.progress:SetPoint("TOPLEFT", f.icon, "TOPRIGHT", 0, 0)
    f.progress:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -5, 5)
    f.progress:SetMinMaxValues(0, 100)
    
    f.lag = CreateFrame("Frame", nil, f.progress)
    f.lag:SetPoint("TOPRIGHT", f.progress, "TOPRIGHT", 0, 0)
    f.lag:SetWidth(f.icon:GetHeight())
    f.lag:SetHeight(f.icon:GetHeight())
    f.lag.texture = f.lag:CreateTexture()
    f.lag.texture:SetAllPoints(f.lag)
    f.lag.texture:SetTexture("Interface\\AddOns\\epsCastbars\\foreground_healthbar")
    f.lag.texture:SetVertexColor(0, 122/255, 0, .5)
    if(unit ~= "player") then f.lag.texture:Hide() end
    
    f.name = f.progress:CreateFontString(nil, "OVERLAY")
    f.name:SetPoint("LEFT", f.progress, "LEFT", 5, 0)
    f.name:SetFont(unpack(epsCore.art.font.default))
    f.name:SetTextColor(1, 1, 1)
    f.name:SetJustifyH("LEFT")
    f.name:SetShadowOffset(1, -1)
    
    f.time = f.lag:CreateFontString(nil, "OVERLAY")
    f.time:SetPoint("RIGHT", f.progress, "RIGHT", -5, 0)
    f.time:SetFont(unpack(epsCore.art.font.default))
    f.time:SetTextColor(1, 1, 1)
    f.time:SetJustifyH("RIGHT")
    f.time:SetShadowOffset(1, -1)
    
    f:RegisterEvent("UNIT_SPELLCAST_START")
    f:RegisterEvent("UNIT_SPELLCAST_STOP")
    f:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
    f:RegisterEvent("UNIT_SPELLCAST_FAILED")
    f:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
    f:RegisterEvent("UNIT_SPELLCAST_DELAYED")
    f:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
    f:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
    f:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
    
    if(unit ~= "player" and unit ~= "target") then
        f:RegisterEvent("UNIT_TARGET")
    elseif(unit == "player") then
        f:RegisterEvent("PLAYER_TARGET_CHANGED")
        f:RegisterEvent("PLAYER_FOCUS_CHANGED")
    end
    f:SetScript("OnEvent", function(self, e)
        if(e == "PLAYER_TARGET_CHANGED") then
            if(epsCastbars:IsCasting("target")) then
                epsCastbars:Start("target")
            else
                epsCastbars:FadeOut("target")
            end
        elseif(e == "PLAYER_FOCUS_CHANGED") then
            if(epsCastbars:IsCasting("focus")) then
                epsCastbars:Start("focus")
            else
                epsCastbars:FadeOut("focus")
            end
        elseif(e == "UNIT_TARGET" and arg1 == self.unit) then
            if(epsCastbars:IsCasting(self.unit)) then
                epsCastbars:Start(self.unit)
            else
                epsCastbars:FadeOut(self.unit)
            end
        elseif(arg1 == self.unit) then
            if(e == "UNIT_SPELLCAST_START" or e == "UNIT_SPELLCAST_CHANNEL_START") then
                epsCastbars:Start(self.unit)
            elseif(e == "UNIT_SPELLCAST_FAILED" or e == "UNIT_SPELLCAST_INTERRUPTED" or e == "UNIT_SPELLCAST_STOP" or e == "UNIT_SPELLCAST_CHANNEL_STOP") then
                epsCastbars:Stop(self.unit)
            end
        end
    end)
    f:SetScript("OnUpdate", function(self, elapsed)
        if(self.fade <= 0) then
            local time = GetTime()
            local milliTime = time*1000
            
            local percent = ((milliTime-self.startTime)/self.duration)*100
            local timeDone = milliTime-self.startTime
            local lag = select(3, GetNetStats())
            local progressText = epsCastbars:round(timeDone/1000, 1)
            
            self.progress:SetValue(percent)
            
            self.lag:SetWidth((self.progress:GetWidth()/100)*(lag/self.duration*100))
            if(self.duration ~= self.initialDuration) then
                self.time:SetFormattedText("|cffdd0000-%s  |cffffffff%s|r", epsCastbars:round((self.duration-self.initialDuration)/1000, 1), progressText)
            else
                self.time:SetFormattedText("|cffffffff%s|r", progressText)
            end
            
            if(percent >= 100) then
                epsCastbars:FadeOut(self.unit)
            end
        else
            if(self:GetAlpha() <= 0) then
                self:Hide()
            else
                self.fadeAlpha = self.fadeAlpha*.9
                self:SetAlpha(self.fadeAlpha)
            end
        end
    end)
    
    return f
end